#include <fits.h>
#include "acspsfdata.h"

ftype fits,pam;
//ftype flash;
double RN,EXP,DMIN,DMAX,EPOCH,FIXED_ET=-1.;
int MASKCR=1;
//char FLASHDIR[161]="";

int ACStype(ftype *f) {
   int i;
   //undrizzled WFC
   if (fits.Next==6) {
      for (i=0;i<6 && f->ext[i].X==4096 && f->ext[i].Y==2048 && f->ext[i].Z==1;i++);
      if (i==6) return 1;
   }
   if (fits.Next==3) {
      //undrizzled HRC
      for (i=0;i<3 && f->ext[i].X==1024 && f->ext[i].Y==1024 && f->ext[i].Z==1;i++);
      if (i==3) return 3;
      //drizzled
      for (i=1;i<3 && f->ext[i].X==f->ext[0].X && f->ext[i].Y==f->ext[0].Y && (i==2 || f->ext[i].Z==f->ext[0].Z);i++);
      if (i==3 && f->ext[0].Z==1) {
	 //drizzled WFC
	 if (f->ext[0].X>=4096 && f->ext[0].Y>=4096) return 2;
	 //drizzled HRC
	 if (f->ext[0].X>=1024 && f->ext[0].Y>=1024 && f->ext[0].X<2000 && f->ext[0].Y<2000) return 4;
      }
   }
   return 0;
}

void ACSgetcards(int ext,int MODE,int drz) {
   int x,y;
   if (strcmp(getcardval(&(fits.img),"FILETYPE",1),"SCI") || strcmp(getcardval(&(fits.img),"TELESCOP",1),"HST") || strcmp(getcardval(&(fits.img),"INSTRUME",1),"ACS")) {printf("**Format error (filetype,telescop,instrume)\n"); exit(-1);}
   if (MODE==0) {
      if (!strcmp(getcardval(&(fits.img),"CCDAMP",1),"A")) RN=atof(getcardval(&(fits.img),"READNSEA",1));
      else if (!strcmp(getcardval(&(fits.img),"CCDAMP",1),"B")) RN=atof(getcardval(&(fits.img),"READNSEB",1));
      else if (!strcmp(getcardval(&(fits.img),"CCDAMP",1),"C")) RN=atof(getcardval(&(fits.img),"READNSEC",1));
      else if (!strcmp(getcardval(&(fits.img),"CCDAMP",1),"D")) RN=atof(getcardval(&(fits.img),"READNSED",1));
      else {printf("**Format error (ccdamp)\n"); exit(-1);}
   }
   else if (MODE==1) {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.5*(atof(getcardval(&(fits.img),"READNSEC",1))+atof(getcardval(&(fits.img),"READNSED",1)));
   }
   else if (MODE==2) {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.5*(atof(getcardval(&(fits.img),"READNSEA",1))+atof(getcardval(&(fits.img),"READNSEB",1)));
   }
   else {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.25*(atof(getcardval(&(fits.img),"READNSEA",1))+atof(getcardval(&(fits.img),"READNSEB",1))+atof(getcardval(&(fits.img),"READNSEC",1))+atof(getcardval(&(fits.img),"READNSED",1)));
   }
   if (FIXED_ET>0) EXP=FIXED_ET;
   else EXP=atof(getcardval(&(fits.img),"EXPTIME",1));
   if (drz) {
      DMIN=DMAX=fits.ext[ext].data[0][0][0];
      for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) {
	 if (fits.ext[ext].data[0][y][x]<DMIN) DMIN=fits.ext[ext].data[0][y][x];
	 else if (fits.ext[ext].data[0][y][x]>DMAX) DMAX=fits.ext[ext].data[0][y][x];
      }
   }
   else {
      DMIN=atof(getcardval(fits.ext+ext,"GOODMIN",1))-1.;
      DMAX=atof(getcardval(fits.ext+ext,"GOODMAX",1))+1.;
   }
   EPOCH=0.5*(atof(getcardval(&(fits.img),"EXPSTART",1))+atof(getcardval(&(fits.img),"EXPEND",1)));
   return;
}

//ignoring type 64 (warm pixel) on assumption it is fixed OK
void ACSmask(ftype*f,int ext,int drz) {
   int x,y,dq,i;
   for (y=0;y<f->ext[ext].Y;y++) for (x=0;x<f->ext[ext].X;x++) if (!drz) {
      dq=(int)(f->ext[ext+2].data[0][y][x]+0.5);
      if (dq&2304) f->ext[ext].data[0][y][x]=DMAX+1;
      else if (dq&5823) f->ext[ext].data[0][y][x]=DMIN-1;
      else if ((dq&8192) && MASKCR) f->ext[ext].data[0][y][x]=DMIN-1;
   }
   else {
      dq=0;
      for (i=0;i<f->ext[ext].Z && !dq;i++) if ((int)(f->ext[ext+2].data[i][y][x]+0.5)!=0) dq=1;
      if (dq==0) f->ext[ext].data[0][y][x]=DMIN-1;
   }
   return;
}

void ACSpixcorr(int ext,char*pamfn,double mult) {
   int x,y;
   double DMIN1,DMAX1;
   char fn[321];

   if (DMAX>0.) DMAX1=1.5*DMAX;
   else DMAX1=0.;
   if (DMIN<0.) DMIN1=1.5*DMIN;
   else DMIN1=0.;
   sprintf(fn,"%s/acs/data/%s",BASEDIR,pamfn);
   readfits(fn,&pam,0);
   if (pam.img.X!=fits.ext[ext].X || pam.img.Y!=fits.ext[ext].Y || pam.img.Z!=fits.ext[ext].Z) {
      printf("Error in PAM file %s\n",pamfn);
      return;
   }
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) if (fits.ext[ext].data[0][y][x]>=DMIN && fits.ext[ext].data[0][y][x]<=DMAX) {
      fits.ext[ext].data[0][y][x]*=pam.img.data[0][y][x]/mult;
      fits.ext[ext+1].data[0][y][x]*=pam.img.data[0][y][x]/mult;
   }
   else if (fits.ext[ext].data[0][y][x]>DMAX) fits.ext[ext].data[0][y][x]=DMAX1+1;
   else fits.ext[ext].data[0][y][x]=DMIN1-1;
   freefits(&pam);
   DMIN=DMIN1;
   DMAX=DMAX1;
   return;
}

void ACSdrzpixcorr(int ext,double mult) {
   int x,y;

   DMAX*=EXP/mult;
   DMIN*=EXP/mult;
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) {
      fits.ext[ext].data[0][y][x]*=EXP/mult;
      fits.ext[ext+1].data[0][y][x]*=EXP/mult;
   }
   return;
}

/*
void ACStrim(int ext,int xf,int yf) {
   int dx=0,dy=0,y,xsize;
   imgtype tmp;
   char str[81],*ptr;

   strcpy(str,getcardval(flash.ext+ext,"LTV1",0));
   if (str[0]) {
      dx=strtol(str,&ptr,10);
      if (ptr==str || (*ptr && strcmp(ptr,".0")) || dx<0 || dx+xf>flash.ext[ext].X) {printf("**Illegal X shift %s\n",str); exit(-1);}
   }
   strcpy(str,getcardval(flash.ext+ext,"LTV2",0));
   if (str[0]) {
      dy=strtol(str,&ptr,10);
      if (ptr==str || (*ptr && strcmp(ptr,".0")) || dy<0 || dy+yf>flash.ext[ext].Y) {printf("**Illegal Y shift %s\n",str); exit(-1);}
   }
   tmp=allocimg(xf,yf,1);
   xsize=sizeof(float)*xf;
   for (y=0;y<yf;y++) memcpy(tmp[0][y],flash.ext[ext].data[0][y+dy]+dx,xsize);
   freeimg(flash.ext[ext].data,flash.ext[ext].X,flash.ext[ext].Y,flash.ext[ext].Z);
   flash.ext[ext].data=tmp;
   flash.ext[ext].X=xf;
   flash.ext[ext].Y=yf;
   return;
}

int loadflash(void) {
   char str[161],*ptr,fcur[81],fn[321];
   int ext,x,y,dq;
   double tflash;

   strcpy(str,getcardval(&(fits.img),"FLASHDUR",1));
   if (!strcmp(str,"0.0")) return 0;
   tflash=strtod(str,&ptr);
   if (ptr==str || *ptr) {printf("**Format error: FLASHDUR\n"); exit(-1);}
   strcpy(str,getcardval(&(fits.img),"FLSHCORR",1));
   if (!strcmp(str,"OMIT")) return 0;
   if (strcmp(str,"COMPLETE")) {printf("**Format error: FLSHCORR\n"); exit(-1);}
   strcpy(fcur,getcardval(&(fits.img),"FLASHCUR",1));
   if (strcmp(fcur,"LOW") && strcmp(fcur,"MED") && strcmp(fcur,"HIGH")) {printf("**Format error: FLASHCUR\n"); exit(-1);}
   strcpy(str,getcardval(&(fits.img),"FLASHSTA",1));
   if (strcmp(str,"SUCCESSFUL")) {printf("**Format error: FLASHSTA\n"); exit(-1);}
   strcpy(str,getcardval(&(fits.img),"FLSHFILE",1));
   if (strncmp(str,"jref$",5)) {printf("**Format error: FLSHFILE\n"); exit(-1);}
   sprintf(fn,"%s%s",FLASHDIR,str+5);
   readfits(fn,&flash,0);
   if (strcmp(getcardval(&(flash.img),"FLASHCUR",1),fcur)) {printf("**Flash/Data mismatch: FLASHCUR\n"); exit(-1);}
   strcpy(str,getcardval(&(flash.img),"FLSHCORR",1));
   if (strcmp(str,"OMIT")) {printf("**Flash format error: FLSHCORR\n"); exit(-1);}
   strcpy(str,getcardval(&(flash.img),"FLASHSTA",1));
   if (strcmp(str,"SUCCESSFUL")) {printf("**Flash format error FLASHSTA\n"); exit(-1);}
   if (fits.Next!=flash.Next) {printf("**Flash format error: Next\n"); exit(-1);}
   for (ext=0;ext<fits.Next;ext+=3) {
      for (y=0;y<flash.ext[ext].Y;y++) for (x=0;x<flash.ext[ext].X;x++) {
	 dq=(int)(flash.ext[ext+2].data[0][y][x]+0.5);
	 if (dq) {printf("**Flash DQ error: nonzero value (%d)\n",dq); exit(-1);}
	 flash.ext[ext].data[0][y][x]*=tflash;
      }
      if (flash.ext[ext].X==4144 && flash.ext[ext].Y==2068) ACStrim(ext,4096,2048);
      else if (flash.ext[ext].X==1062 && flash.ext[ext].Y==1044) ACStrim(ext,1024,1024);
      else {printf("**Flash size error: %dx%d\n",flash.ext[ext].X,flash.ext[ext].Y); exit(-1);}
      if (ext) memcpy(flash.ext+ext/3,flash.ext+ext,sizeof(imtype));
      freeim(flash.ext+ext+1);
      freeim(flash.ext+ext+2);
   }
   flash.Next/=3;
   return 1;
}

void ACSflashcorr(int ext) {
   double DMIN1=0.,DMAX1=0.;
   int x,y;

   if (flash.ext[ext].X!=fits.ext[ext].X || flash.ext[ext].Y!=fits.ext[ext].Y || flash.ext[ext].Z!=fits.ext[ext].Z) {printf("**Error in FLASH file (size mismatch)\n"); exit(-1);}
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) {
      if (flash.ext[ext].data[0][y][x]>DMAX1) DMAX1=flash.ext[ext].data[0][y][x];
      else if (flash.ext[ext].data[0][y][x]<DMIN1) DMIN1=flash.ext[ext].data[0][y][x];
   }
   DMAX1+=DMAX+1.;
   DMIN1+=DMIN-1.;
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) if (fits.ext[ext].data[0][y][x]>=DMIN && fits.ext[ext].data[0][y][x]<=DMAX) fits.ext[ext].data[0][y][x]+=flash.ext[ext].data[0][y][x];
   else if (fits.ext[ext].data[0][y][x]>DMAX) fits.ext[ext].data[0][y][x]=DMAX1+1;
   else fits.ext[ext].data[0][y][x]=DMIN1-1;
   DMAX=DMAX1;
   DMIN=DMIN1;
   return;
}
*/

void ACSsetcards(int ext) {
   int x,y;
   chiptype data,noise;
   double I=0,X=0,Y=0,XX=0,XY=0,tGN,tRN;

   data=fits.ext[ext].data[0];
   noise=fits.ext[ext+1].data[0];
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) if (data[y][x]>=DMIN && data[y][x]<=DMAX) {
      I++;
      Y+=noise[y][x]*noise[y][x];
      if (data[y][x]>0.) {
	 X+=data[y][x];
	 XX+=data[y][x]*data[y][x];
	 XY+=data[y][x]*noise[y][x]*noise[y][x];
      }
   }
   tGN=(XX*I-X*X)/(XY*I-X*Y);
   tRN=sqrt((XX*Y-X*XY)/(XX*I-X*X))*tGN;
   printf("%f %f\n",tGN,tRN);
   insertcards(fits.ext+ext,1.,RN,EXP,DMIN,DMAX,EPOCH,0.0);
   freeim(fits.ext+ext+1);
   freeim(fits.ext+ext+2);
   fits.Next-=2;
   for (x=ext+1;x<fits.Next;x++) memcpy(fits.ext+x,fits.ext+x+2,sizeof(imtype));
   return;
}

int main(int argc,char**argv) {
   int i,tp,psize;
   //int fcorr=0;

   if (argc<2) {
      printf("Usage: %s <<-flags>> <fits files>\n",*argv);
      printf(" -keepcr      leaves fixed cosmic ray pixels in image\n");
      printf(" -exptime=#   overrides exposure time; needed for _drz images\n");
      //printf(" -flashdir=X  sets directory prefix for flash correction files\n");
      return 1;
   }
   for (i=1;i<argc;i++) if (!strcasecmp(argv[i],"-KEEPCR")) MASKCR=0;
   else if (!strncasecmp(argv[i],"-EXPTIME=",9)) FIXED_ET=atof(argv[i]+9);
   //else if (!strncasecmp(argv[i],"-FLASHDIR=",10)) strcpy(FLASHDIR,argv[i]+10);
   else {
      psize=1;
      readfits(argv[i],&fits,1);
      tp=ACStype(&fits);
      if (fits.Next>0 && strcmp(getcardval(fits.ext,"DOL_ACS",0),"")) printf("%s already run through acsmask\n",argv[i]);
      else if (tp<1 || tp>4) printf("%s is not an ACS fits file\n",argv[i]);
      else {
	 //fcorr=loadflash();
	 if (tp==1) { // FLT or CRJ WFC image
	    ACSgetcards(0,2,0);
	    ACSmask(&fits,0,0);
	    ACSpixcorr(0,"wfc2_pam.fits",acs_ctmult[2]);
	    //if (fcorr) ACSflashcorr(0);
	    ACSsetcards(0);
	    ACSgetcards(1,1,0);
	    ACSmask(&fits,1,0);
	    ACSpixcorr(1,"wfc1_pam.fits",acs_ctmult[1]);
	    //if (fcorr) ACSflashcorr(1);
	    ACSsetcards(1);
	    addcard(fits.ext,"DOL_ACS =                    2 / DOLPHOT ACS tag                                ");
	    addcard(fits.ext+1,"DOL_ACS =                    1 / DOLPHOT ACS tag                                ");
	    fits.Next=2;
	 }
	 else if (tp==2) { // DRZ WFC image
	    //if (fcorr) printf("%s is drizzled and flashed\n",argv[i]);
	    //else {
	       ACSgetcards(0,-1,1);
	       ACSmask(&fits,0,1);
	       ACSdrzpixcorr(0,acs_ctmult[1]);
	       ACSsetcards(0);
	       fits.Next=1;
	       addcard(fits.ext,"DOL_ACS =                   -2 / DOLPHOT ACS tag                                ");
	    //}
	 }
	 else if (tp==3) { // FLT or CRJ HRC image
	    ACSgetcards(0,0,0);
	    ACSmask(&fits,0,0);
	    ACSpixcorr(0,"hrc_pam.fits",acs_ctmult[0]);
	    //if (fcorr) ACSflashcorr(0);
	    ACSsetcards(0);
	    fits.Next=1;
	    addcard(fits.ext,"DOL_ACS =                    0 / DOLPHOT ACS tag                                ");
	 }
	 else if (tp==4) { // DRZ HRC image
	    //if (fcorr) printf("%s is drizzled and flashed\n",argv[i]);
	    //else {
	       ACSgetcards(0,0,1);
	       ACSmask(&fits,0,1);
	       ACSdrzpixcorr(0,acs_ctmult[0]);
	       ACSsetcards(0);
	       fits.Next=1;
	       addcard(fits.ext,"DOL_ACS =                   -1 / DOLPHOT ACS tag                                ");
	    //}
	 }
	 writefits(argv[i],&fits,1);
      }
      //if (fcorr) freefits(&flash);
      freefits(&fits);
   }
   return 0;
}
