#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fits.h>
#include "acspsfdata.h"
#include "acsfilters.h"
#include "acsdistort.h"

typedef double apsftype[3][6];
extern apsftype *apsf;
extern imtype *datahd,*dataim;
extern int Timg,Nimg;
extern int rphot,RPSF,SubPixel,FPSF,EPSF;
extern double RAper,MinS,MaxS,MaxE,PSFStep,Zero,*apsize;
extern int PSFsol;

typedef int int2[2];
int2 *acspsflu=NULL;
float********acspsflib=NULL;
int DRIZZLE_BASE=0,ACS_RPSF=34,FlagMask=0;

void acsinitparam(void) {
   FPSF=4;
   SubPixel=1;
   EPSF=0;
   PSFsol=-1;
   PSFStep=0.;
   Zero=0.;
   MinS=0.2;
   MaxS=15.;
   MaxE=0.1;
   if (RPSF>=acs_rpsf[0]) {RPSF=acs_rpsf[0]-1; printf("Lowering RPSF to %d\n",RPSF);}
   if (RAper>acs_rpsf[0]-1.) {RAper=acs_rpsf[0]-1; printf("Lowering RAper to %d\n",acs_rpsf[0]-1);}
   rphot=(int)(RAper+0.9999);
   return;
}

void acsinitpsf(void) {
   int img,i,j,y1,x1,y2,x2,y3,n2,n3,n3skip;
   char str[81],*ptr;
   FILE *f;
   float***** ptr1;
   float**** ptr2;
   float*** ptr3;
   float** ptr4;
   float* ptr5;

   if (NFILTERS<0) initfilters();
   if (acspsflib==NULL) {
      acspsflib=(float********)calloc(sizeof(float*******),3);
      if (!acspsflib) merr();
      for (i=0;i<3;i++) {
	 acspsflib[i]=(float*******)calloc(sizeof(float******),NFILTERS);
	 if (!acspsflib[i]) merr();
      }
   }
   if (acspsflu==NULL) {
      acspsflu=(int2*)calloc(sizeof(int2),Timg);
      if (!acspsflu) merr();
   }
   for (i=0;i<3;i++) for (j=0;j<NFILTERS;j++) acspsflib[i][j]=NULL;
   for (img=0;img<Timg;img++) {
      strcpy(str,getcardval(dataim+img,"DOL_ACS",0));
      acspsflu[img][0]=strtol(str,&ptr,10);
      if (ptr==str) {
	 printf("**Image %d has not been preprocessed with acsmask; cannot proceed\n",img+1);
	 exit(-1);
      }
      else if (acspsflu[img][0]==-1 || (acspsflu[img][0]==-2 && img<Nimg)) {
	 printf("**Image %d has been drizzled; cannot run photometry\n",img+1);
	 exit(-1);
      }
      else if (acspsflu[img][0]<-2 || acspsflu[img][0]>2) {
	 printf("**Image %d's chip cannot be identified; please report bug.\n",img+1);
	 exit(-1);
      }
      else {
	 if (acspsflu[img][0]==-2) {
	    acspsflu[img][0]=1;
	    DRIZZLE_BASE=1;
	 }
	 strcpy(str,getcardval(datahd+img,"FILTER1",0));
	 if (acspsflu[img][0]>0 && !strcmp(str,"CLEAR1L")) strcpy(str,getcardval(datahd+img,"FILTER2",0));
	 else if (acspsflu[img][0]==0 && !strcmp(str,"CLEAR1S")) strcpy(str,getcardval(datahd+img,"FILTER2",0));
	 acspsflu[img][1]=findfilt(str);
	 //printf("Image %d: cm=%d, filt=%d (%s)\n",img+1,acspsflu[img][0],acspsflu[img][1],filters[acspsflu[img][1]].name);
      }
      if (RPSF>=acs_rpsf[acspsflu[img][0]]) {printf("ERROR: RPSF must be less than %d for ACS/WFC data\n",acs_rpsf[acspsflu[img][0]]); exit(-1);}
      if (RAper>acs_rpsf[acspsflu[img][0]]-1.) {printf("ERROR: RAper must be no more than %d for ACS/WFC data\n",acs_rpsf[acspsflu[img][0]]-1); exit(-1);}
   }
   for (img=0;img<Timg;img++) {
      i=acspsflu[img][0];
      apsf[img][0][0]=1.;
      apsf[img][1][0]=1.;
      apsf[img][2][0]=0.;
      if (i==0) apsize[img]=11.;
      else apsize[img]=10.;
      for (i=1;i<5;i++) apsf[img][0][i]=apsf[img][1][i]=apsf[img][2][i]=0.;
   }
   ACS_RPSF=RPSF;
   if (ACS_RPSF<rphot) ACS_RPSF=rphot;
   for (img=0;img<Timg;img++) if (acspsflib[i=acspsflu[img][0]][j=acspsflu[img][1]]==NULL) {
      sprintf(str,"%s/acs/data/%s.%s.psf",BASEDIR,filters[acspsflu[img][1]].name,acs_cn[i]);
      if ((f=fopen(str,"rb"))==NULL) {
	 printf("Cannot open %s\n",str);
	 exit(-1);
      }
      n2=2*acs_n2psf[i]+1;
      n3=2*ACS_RPSF+1;
      n3skip=acs_rpsf[i]-ACS_RPSF;
      acspsflib[i][j]=(float******)calloc(sizeof(float*****),acs_nypsfpos[i]);
      ptr1=(float*****)calloc(sizeof(float****),acs_nypsfpos[i]*acs_nxpsfpos[i]);
      ptr2=(float****)calloc(sizeof(float***),acs_nypsfpos[i]*acs_nxpsfpos[i]*n2);
      ptr3=(float***)calloc(sizeof(float**),acs_nypsfpos[i]*acs_nxpsfpos[i]*n2*n2);
      ptr4=(float**)calloc(sizeof(float*),acs_nypsfpos[i]*acs_nxpsfpos[i]*n2*n2*n3);
      ptr5=(float*)calloc(sizeof(float),acs_nypsfpos[i]*acs_nxpsfpos[i]*n2*n2*n3*n3);
      if (!acspsflib[i][j] || !ptr1) merr();
      for (y1=0;y1<acs_nypsfpos[i];y1++) {
	 acspsflib[i][j][y1]=ptr1;
	 ptr1+=acs_nxpsfpos[i];
	 for (x1=0;x1<acs_nxpsfpos[i];x1++) {
	    acspsflib[i][j][y1][x1]=ptr2+acs_n2psf[i];
	    ptr2+=n2;
	    for (y2=-acs_n2psf[i];y2<=acs_n2psf[i];y2++) {
	       acspsflib[i][j][y1][x1][y2]=ptr3+acs_n2psf[i];
	       ptr3+=n2;
	       for (x2=-acs_n2psf[i];x2<=acs_n2psf[i];x2++) {
		  acspsflib[i][j][y1][x1][y2][x2]=ptr4+ACS_RPSF;
		  ptr4+=n3;
		  if (n3skip) fseek(f,4*n3skip*(2*acs_rpsf[i]+1),SEEK_CUR);
		  for (y3=-ACS_RPSF;y3<=ACS_RPSF;y3++) {
		     acspsflib[i][j][y1][x1][y2][x2][y3]=ptr5+ACS_RPSF;
		     ptr5+=n3;
		     if (n3skip) fseek(f,4*n3skip,SEEK_CUR);
		     ffread(acspsflib[i][j][y1][x1][y2][x2][y3]-ACS_RPSF,4,n3,f);
		     if (n3skip) fseek(f,4*n3skip,SEEK_CUR);
		  }
		  if (n3skip) fseek(f,4*n3skip*(2*acs_rpsf[i]+1),SEEK_CUR);
	       }
	    }
	 }
      }
      fclose(f);
   }
   return;
}

void acsfreepsf(void) {
   int i,j;
   for (i=0;i<3;i++) for (j=0;j<NFILTERS;j++) if (acspsflib[i][j]) {
      free(acspsflib[i][j][0][0][-acs_n2psf[i]][-acs_n2psf[i]][-ACS_RPSF]-ACS_RPSF);
      free(acspsflib[i][j][0][0][-acs_n2psf[i]][-acs_n2psf[i]]-ACS_RPSF);
      free(acspsflib[i][j][0][0][-acs_n2psf[i]]-acs_n2psf[i]);
      free(acspsflib[i][j][0][0]-acs_n2psf[i]);
      free(acspsflib[i][j][0]);
      free(acspsflib[i][j]);
      acspsflib[i][j]=NULL;
   }
   return;
}

extern int lastpsftype;
extern float **psf;

/*
void calcacspsf(int img,float x,float y,int r) {
   int i,j,y1,x1,y2,x2,yy,xx;
   float mx1,my1,imx1,imy1;
   float mx2,my2,imx2,imy2;
   static int first=1,lastr=0,lastimg=0;
   static float lastx=0,lasty=0;

   if (!first && lastpsftype==1 && img==lastimg && x==lastx && y==lasty && r<=lastr) return;
   first=0;
   lastpsftype=1;
   lastimg=img;
   lastx=x;
   lasty=y;
   lastr=r;
   i=acspsflu[img][0];
   j=acspsflu[img][1];
   //printf("%s/%s PSF at %f,%f; r=%d:\n",acs_cn[i],filters[j].name,x,y,r);
   y1=(int)(y-128)/256; if (y1<0) y1=0; if (y1>=acs_nypsfpos[i]-1) y1=acs_nypsfpos[i]-2;
   imy1=(y-128-256*y1)/256.; my1=1-imy1;
   x1=(int)(x-128)/256; if (x1<0) x1=0; if (x1>=acs_nxpsfpos[i]-1) x1=acs_nxpsfpos[i]-2;
   imx1=(x-128-256*x1)/256.; mx1=1-imx1;
   y=(y-(int)(y+50)+49.5)*acs_sub[i];
   y2=(int)(y+acs_sub[i])-acs_sub[i];
   imy2=y-y2; my2=1-imy2;
   x=(x-(int)(x+50)+49.5)*acs_sub[i];
   x2=(int)(x+acs_sub[i])-acs_sub[i];
   imx2=x-x2; mx2=1-imx2;
   //printf("y1=%d, my1=%f; x1=%d, mx1=%f\n",y1,my1,x1,mx1);
   //printf("y2=%d, my2=%f; x2=%d, mx2=%f\n",y2,my2,x2,mx2);
   for (yy=-r;yy<=r;yy++) for (xx=-r;xx<=r;xx++) {
      psf[yy][xx]=0.;
      psf[yy][xx]+=(acspsflib[i][j][y1][x1][y2][x2][yy][xx]*mx2*my2+acspsflib[i][j][y1][x1][y2][x2+1][yy][xx]*imx2*my2+acspsflib[i][j][y1][x1][y2+1][x2][yy][xx]*mx2*imy2+acspsflib[i][j][y1][x1][y2+1][x2+1][yy][xx]*imx2*imy2)*mx1*my1;
      psf[yy][xx]+=(acspsflib[i][j][y1][x1+1][y2][x2][yy][xx]*mx2*my2+acspsflib[i][j][y1][x1+1][y2][x2+1][yy][xx]*imx2*my2+acspsflib[i][j][y1][x1+1][y2+1][x2][yy][xx]*mx2*imy2+acspsflib[i][j][y1][x1+1][y2+1][x2+1][yy][xx]*imx2*imy2)*imx1*my1;
      psf[yy][xx]+=(acspsflib[i][j][y1+1][x1][y2][x2][yy][xx]*mx2*my2+acspsflib[i][j][y1+1][x1][y2][x2+1][yy][xx]*imx2*my2+acspsflib[i][j][y1+1][x1][y2+1][x2][yy][xx]*mx2*imy2+acspsflib[i][j][y1+1][x1][y2+1][x2+1][yy][xx]*imx2*imy2)*mx1*imy1;
      psf[yy][xx]+=(acspsflib[i][j][y1+1][x1+1][y2][x2][yy][xx]*mx2*my2+acspsflib[i][j][y1+1][x1+1][y2][x2+1][yy][xx]*imx2*my2+acspsflib[i][j][y1+1][x1+1][y2+1][x2][yy][xx]*mx2*imy2+acspsflib[i][j][y1+1][x1+1][y2+1][x2+1][yy][xx]*imx2*imy2)*imx1*imy1;
   }
   //for (yy=6;yy>=-6;yy--) if (yy>=-r && yy<=r) {for (xx=-6;xx<=6;xx++) if (xx>=-r && xx<=r) printf("%5d ",(int)(psf[yy][xx]*100000+0.5)); printf("\n");} fflush(stdout);
   return;
}
*/

int calcacspsf(int img,float x,float y,int r) {
   int i,j,y1,x1,y2,x2,yy,xx;
   float mx2,my2,imx2,imy2;
   static int first=1,lastr=0,lastimg=0;
   static float lastx=0,lasty=0;
   extern int poffreset;

   if (!first && lastpsftype==1 && img==lastimg && x==lastx && y==lasty && r<=lastr && !poffreset) return 0;
   first=0;lastpsftype=1;lastimg=img;lastx=x;lasty=y;lastr=r;
   i=acspsflu[img][0];
   j=acspsflu[img][1];
   //printf("%d %d %d %f %f %d",img,i,j,x,y,r); fflush(stdout);
   //printf("%s/%s PSF at %f,%f; r=%d:\n",acs_cn[i],filters[j].name,x,y,r);
   if (img==Nimg && DRIZZLE_BASE) {
      y1=4;
      x1=7;
   }
   else {
      y1=(int)y/256; if (y1<0) y1=0; if (y1>=acs_nypsfpos[i]-1) y1=acs_nypsfpos[i]-2;
      x1=(int)x/256; if (x1<0) x1=0; if (x1>=acs_nxpsfpos[i]-1) x1=acs_nxpsfpos[i]-2;
   }
   y-=(int)y; if (y<0) y++;
   y=(y-0.5)*acs_sub[i];
   y2=(int)(y+acs_sub[i])-acs_sub[i];
   imy2=y-y2; my2=1-imy2;
   //printf(" (%f %d %d %f %f)",y,y1,y2,my2,imy2); fflush(stdout);
   x-=(int)x; if (x<0) x++;
   x=(x-0.5)*acs_sub[i];
   x2=(int)(x+acs_sub[i])-acs_sub[i];
   imx2=x-x2; mx2=1-imx2;
   //printf(" (%f %d %d %f %f)\n",x,x1,x2,mx2,imx2); fflush(stdout);
   //printf("y1=%d, my1=%f; x1=%d, mx1=%f\n",y1,my1,x1,mx1);
   //printf("y2=%d, my2=%f; x2=%d, mx2=%f\n",y2,my2,x2,mx2);
   for (yy=-r;yy<=r;yy++) for (xx=-r;xx<=r;xx++) psf[yy][xx]=acspsflib[i][j][y1][x1][y2][x2][yy][xx]*mx2*my2+acspsflib[i][j][y1][x1][y2][x2+1][yy][xx]*imx2*my2+acspsflib[i][j][y1][x1][y2+1][x2][yy][xx]*mx2*imy2+acspsflib[i][j][y1][x1][y2+1][x2+1][yy][xx]*imx2*imy2;
   //for (yy=6;yy>=-6;yy--) if (yy>=-r && yy<=r) {for (xx=-6;xx<=6;xx++) if (xx>=-r && xx<=r) printf("%5d ",(int)(psf[yy][xx]*100000+0.5)); printf("\n");} fflush(stdout);
   return 1;
}

extern double *iGAIN,*iEXP,*iEPOCH,*apcor;
extern void shift(int img,float x0,float y0,float *x,float *y,int dir);
double ACScalcmag(int img,float x0,float y0,float ct0,float bg) {
   int cm;
   float x,y;
   char str[81];
   double m;

   if (NFILTERS<0) initfilters();
   m=-2.5*log10(ct0*apcor[img]/iEXP[img]*acs_ctmult[acspsflu[img][0]])+filters[acspsflu[img][1]].zp[acspsflu[img][0]];
   shift(img,x0,y0,&x,&y,1);
   strcpy(str,getcardval(datahd+img,"NRPTEXP",0));
   if (*str==0) {
      printf("Error: NRPTEXP not found; assuming 1\n");
      cm=1;
   }
   else cm=atoi(str);
   strcpy(str,getcardval(datahd+img,"CRSPLIT",0));
   if (*str==0) printf("Error: CRSPLIT not found; assuming 1\n");
   else cm*=atoi(str);
   m-=ACS_CTE(acspsflu[img][0],x,y,ct0,cm,iGAIN[img],bg,iEPOCH[img]);
   return m;
}

typedef struct{float ct0,ct,sky,m,dm,chi,sn,sh,rnd,r,e,a,b,c,crowd;int flag;} photdatatype;
void ACSoutstar(FILE *of,float x0,float y0,photdatatype*pdata) {
   int i,j,cm;
   static int *fused=NULL;
   static photdatatype*fphot=NULL;
   static double*smag=NULL,*vmag=NULL,*dvmag=NULL;
   float x,y;
   char str[81];

   if (NFILTERS<0) initfilters();
   if (!fphot) {
      fused=(int*)calloc(sizeof(int),NFILTERS);
      fphot=(photdatatype*)calloc(sizeof(photdatatype),NFILTERS);
      smag=(double*)calloc(sizeof(double),NFILTERS);
      vmag=(double*)calloc(sizeof(double),NFILTERS);
      dvmag=(double*)calloc(sizeof(double),NFILTERS);
      if (!fused || !fphot || !smag || !vmag || !dvmag) merr();
   }
   for (j=0;j<Nimg;j++) if (pdata[j].ct>0) {
      pdata[j].m-=2.5*log10(acs_ctmult[acspsflu[j][0]]);
      pdata[j].m+=filters[acspsflu[j][1]].zp[acspsflu[j][0]]-Zero;
      shift(j,x0,y0,&x,&y,1);
      strcpy(str,getcardval(datahd+j,"NRPTEXP",0));
      if (*str==0) {
	 printf("Error: NRPTEXP not found; assuming 1\n");
	 cm=1;
      }
      else cm=atoi(str);
      strcpy(str,getcardval(datahd+j,"CRSPLIT",0));
      if (*str==0) printf("Error: CRSPLIT not found; assuming 1\n");
      else cm*=atoi(str);
      pdata[j].m-=ACS_CTE(acspsflu[j][0],x,y,pdata[j].ct,cm,iGAIN[j],pdata[j].sky,iEPOCH[j]);
   }
   for (i=0;i<NFILTERS;i++) {
      float wt,swt,twt=0,tswt=0,S=0.,SS=0.,is,iss,tcm=0.;
      fphot[i].ct0=fphot[i].ct=fphot[i].sky=fphot[i].dm=fphot[i].chi=fphot[i].sn=fphot[i].sh=fphot[i].rnd=fphot[i].r=fphot[i].e=fphot[i].a=fphot[i].b=fphot[i].c=fphot[i].m=fphot[i].crowd=0.;
      fused[i]=0;
      fphot[i].flag=0;
      for (j=0;j<Nimg;j++) if (acspsflu[j][1]==i) {
	 fused[i]++;
	 if (pdata[j].ct>0 && pdata[j].flag<8 && !(pdata[j].flag&FlagMask)) {
	    is=pdata[j].ct/iEXP[j];
	    iss=is/pdata[j].sn;
	    wt=1./iss/iss;
	    swt=wt*is;
	    twt+=wt;
	    tswt+=swt;
	    fphot[i].ct0+=pdata[j].ct0/iEXP[j]*wt;
	    fphot[i].ct+=pdata[j].ct/iEXP[j]*wt;
	    fphot[i].chi+=pdata[j].chi*pdata[j].chi*wt;
	    fphot[i].sh+=pdata[j].sh*swt;
	    fphot[i].sky+=pdata[j].sky/iEXP[j]*wt;
	    fphot[i].m+=pow(10,-0.4*pdata[j].m)*wt;
	    fphot[i].rnd+=pdata[j].rnd*swt;
	    fphot[i].crowd+=pdata[j].crowd*wt;
	    fphot[i].r+=pdata[j].r*swt;
	    fphot[i].e+=pdata[j].e*swt;
	    fphot[i].a+=pdata[j].a*swt;
	    fphot[i].b+=pdata[j].b*swt;
	    fphot[i].c+=pdata[j].c*swt;
	    fphot[i].flag|=pdata[j].flag;
	    S+=is*wt;
	    SS+=iss*iss*wt*wt;
	 }
	 tcm+=iEXP[j];
      }
      if (twt>0.) {
	 fphot[i].ct0/=twt/tcm;
	 fphot[i].ct/=twt/tcm;
	 fphot[i].chi=sqrt(fphot[i].chi/twt);
	 fphot[i].sky/=twt/tcm;
	 fphot[i].m=-2.5*log10(fphot[i].m/twt);
	 fphot[i].sh/=tswt;
	 fphot[i].rnd/=tswt;
	 fphot[i].crowd/=twt;
	 fphot[i].r/=tswt;
	 fphot[i].e/=tswt;
	 fphot[i].a/=tswt;
	 fphot[i].b/=tswt;
	 fphot[i].c/=tswt;
	 S/=twt;
	 SS=sqrt(SS)/twt;
	 fphot[i].sn=S/SS;
	 fphot[i].dm=1.0857362*SS/S;
      }
      else fphot[i].m=99.999;
      vmag[i]=fphot[i].m;
      dvmag[i]=fphot[i].dm;
   }
   cm=(1+acspsflu[0][0])/2;
   for (j=1;j<Nimg;j++) if (cm!=(1+acspsflu[j][0])/2) {
      printf("Multi-mode transformations not yet supported!\n");
      exit(-1);
   }
   transform(cm,vmag,dvmag,smag);
   for (i=0;i<NFILTERS;i++) if (fused[i]>1) {
      if (fphot[i].ct<999999.5) fprintf(of,"  %8.1f",fphot[i].ct);
      else fprintf(of,"  %8.2e",fphot[i].ct);
      if (fphot[i].sky<99999.95) fprintf(of," %8.2f",fphot[i].sky);
      else if (fphot[i].sky<999999.5) fprintf(of," %8.1f",fphot[i].sky);
      else fprintf(of," %8.1e",fphot[i].sky);
      fprintf(of," %6.3f %6.3f",fphot[i].m,smag[i]);
      if (fphot[i].dm>9.999) fprintf(of," 9.999");
      else fprintf(of," %5.3f",fphot[i].dm);
      fprintf(of," %6.2f %7.1f %6.3f %6.3f %5.3f %6.3f %6.3f %6.3f %6.3f %6.3f %2d",fphot[i].chi,fphot[i].sn,fphot[i].sh,fphot[i].rnd,fphot[i].crowd,fphot[i].r,fphot[i].e,fphot[i].a,fphot[i].b,fphot[i].c,fphot[i].flag);
   }
   for (j=0;j<Nimg;j++) {
      if (pdata[j].ct<999999.5) fprintf(of,"  %8.1f",pdata[j].ct);
      else fprintf(of,"  %8.2e",pdata[j].ct);
      if (pdata[j].sky<99999.95) fprintf(of," %8.2f",pdata[j].sky);
      else if (pdata[j].sky<999999.5) fprintf(of," %8.1f",pdata[j].sky);
      else fprintf(of," %8.1e",pdata[j].sky);
      if (pdata[j].ct<=0) x=pdata[j].m;
      else if (smag[acspsflu[j][1]]>99) x=smag[acspsflu[j][1]];
      else x=pdata[j].m+smag[acspsflu[j][1]]-vmag[acspsflu[j][1]];
      fprintf(of," %6.3f %6.3f",pdata[j].m,x);
      if (pdata[j].dm>9.999) fprintf(of," 9.999");
      else fprintf(of," %5.3f",pdata[j].dm);
      fprintf(of," %6.2f %7.1f %6.3f %6.3f %5.3f %6.3f %6.3f %6.3f %6.3f %6.3f %2d",pdata[j].chi,pdata[j].sn,pdata[j].sh,pdata[j].rnd,pdata[j].crowd,pdata[j].r,pdata[j].e,pdata[j].a,pdata[j].b,pdata[j].c,pdata[j].flag);
   }
   return;
}

double ACSidcfwd0[3][25]={
   {531,512,205.1358,470.1861,-0.07,0.025,0.000030,0.028415,-3.998E-08,2.373E-07,-1.047E-07,9.507E-12,-2.084E-11,6.676E-12,-2.627E-11,2.469E-02,2.804E-03,2.771E-07,-5.868E-08,3.675E-08,6.095E-11,-1.247E-11,3.535E-11,2.432E-11,1.123113},
   {2072,1024,256.4158,237.0618,-2.38,0.050,0.002032,0.049160,1.029E-07,-3.527E-07,4.240E-07,4.461E-12,-2.258E-11,-5.327E-12,-2.187E-11,4.895E-02,2.231E-03,-4.904E-07,2.955E-07,-1.304E-07,-1.571E-11,-1.996E-12,-1.846E-11,3.686E-12,0.9809135},
   {2072,1024,254.2550,301.1133,-1.94,0.050,0.001718,0.049779,9.929E-08,-2.541E-07,4.316E-07,-1.055E-13,-2.606E-11,-2.602E-12,-2.286E-11,5.068E-02,1.583E-03,-3.677E-07,3.068E-07,-8.124E-08,-1.960E-11,-4.799E-12,-2.736E-11,3.714E-12,0.9809135}
};

float acs_apsize(int img,float x,float y) {
   float xc,yc,p2x,p2y,p2,p1x,p1y,p1,th,area;
   int cm=acspsflu[img][0];
   if (x<0) x=0;
   if (y<0) y=0;
   if (cm==0) {
      if (x>1024) x=1024;
      if (y>1024) y=1024;
   }
   else {
      if (x>4096) x=4096;
      if (y>2048) y=2048;
   }
   //compute pixel areas;
   xc=x-ACSidcfwd0[cm][0];
   yc=y-ACSidcfwd0[cm][1];
   p2x=(ACSidcfwd0[cm][7]+ACSidcfwd0[cm][9]*yc+2*ACSidcfwd0[cm][10]*xc+ACSidcfwd0[cm][12]*yc*yc+2*ACSidcfwd0[cm][13]*xc*yc+3*ACSidcfwd0[cm][14]*xc*xc)/ACSidcfwd0[cm][5]; //dX/dx
   p2y=(ACSidcfwd0[cm][6]+2*ACSidcfwd0[cm][8]*yc+ACSidcfwd0[cm][9]*xc+3*ACSidcfwd0[cm][11]*yc*yc+2*ACSidcfwd0[cm][12]*xc*yc+ACSidcfwd0[cm][13]*xc*xc)/ACSidcfwd0[cm][5]; //dY/dx
   p2=sqrt(p2x*p2x+p2y*p2y);
   p1x=(ACSidcfwd0[cm][16]+ACSidcfwd0[cm][18]*yc+2*ACSidcfwd0[cm][19]*xc+ACSidcfwd0[cm][21]*yc*yc+2*ACSidcfwd0[cm][22]*xc*yc+3*ACSidcfwd0[cm][23]*xc*xc)/ACSidcfwd0[cm][5]; //dX/dy
   p1y=(ACSidcfwd0[cm][15]+2*ACSidcfwd0[cm][17]*yc+ACSidcfwd0[cm][18]*xc+3*ACSidcfwd0[cm][20]*yc*yc+2*ACSidcfwd0[cm][21]*xc*yc+ACSidcfwd0[cm][22]*xc*xc)/ACSidcfwd0[cm][5]; //dY/dy
   p1=sqrt(p1x*p1x+p1y*p1y);
   th=acos((p1x*p2x+p1y*p2y)/p1/p2);
   area=p1*p2*sin(th);
   if (cm==0) return 12./sqrt(area);
   return 10./sqrt(area);
}

void ACSshift(int img,float*x,float*y) {
   double xx,yy;
   xx=*x; yy=*y;
   if (acspsflu[img][0]>=0) ACSfwddistort(acspsflu[img][0],acspsflu[img][1],&xx,&yy);
   *x=xx; *y=yy;
   return;
}

void ACSunshift(int img,float*x,float*y) {
   double xx,yy;
   xx=*x; yy=*y;
   if (acspsflu[img][0]>=0) ACSrevdistort(acspsflu[img][0],acspsflu[img][1],&xx,&yy);
   *x=xx; *y=yy;
   return;
}

extern FILE *finfo;
void writeacsinfo(void) {
   int img;
   fprintf(finfo,"* ACS-specific info\n");
   for (img=0;img<Nimg;img++) {
      fprintf(finfo,"* %s %d %f\n",filters[acspsflu[img][1]].name,acspsflu[img][0],iEXP[img]);
      if (fabs(iGAIN[img]-1.)>0.001) printf("ERROR: All ACS data should have gain of 1\n");
   }
   return;
}

int*ffused=NULL;
double*fakem0;
void ACSreadfakemag(FILE*f) {
   int i,img;

   if (NFILTERS<0) initfilters();
   if (ffused==NULL) {
      ffused=(int*)calloc(sizeof(int),NFILTERS);
      fakem0=(double*)calloc(sizeof(double),NFILTERS);
      if (!ffused || !fakem0) merr();
      for (img=0;img<Nimg;img++) ffused[acspsflu[img][1]]++;
   }
   for (i=0;i<NFILTERS;i++) if (ffused[i]) fscanf(f,"%lf",fakem0+i);
   return;
}

void ACSfixfakemag(int img,float x0,float y0,double*ct0,float*bg) {
   int i;
   double dm;

   for (i=0;i<3;i++) {
      dm=ACScalcmag(img,x0,y0,ct0[img],bg[img])-fakem0[acspsflu[img][1]];
      ct0[img]*=pow(10,0.4*dm);
   }
   return;
}

